/* -*- Mode: C; indent-tabs-mode: nil; c-basic-offset: 2; tab-width: 2 -*- */
/*
    Multiplication Puzzle
    Copyright (C) 2004-2008 Michael Terry <mike@mterry.name>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

using GLib;

public class GtkMult : Gtk.Window
{
  public static int dpi;
  
  public MultPuzzle puzzle { get; private set; }
  public bool cheated { get; private set; }
  public Gtk.Statusbar status { get; private set; }
  
  public static const double ASPECT = 8.0/6.0;
  
  construct
  {
    vbox = new Gtk.VBox (false, 0);
    status = null;
    score = new Gtk.Label ("");
    canvas = null;
    status_timeout_id = 0;
    solve_action = null;
    
    // Convert pixel space to desired point size
    Value dpiValue = Value (typeof (int));
    get_screen ().get_setting ("gtk-xft-dpi", dpiValue);
    this.dpi = dpiValue.get_int () / Pango.SCALE;
    
    var height = 4*dpi;
    set_default_size((int)(height*ASPECT), height);
    
    Gtk.VBox vb = new Gtk.VBox (false, 0);
    status_box = new Gtk.HBox (false, 0);
    Gtk.EventBox eb = new Gtk.EventBox ();

    vb.pack_start (setup_menu (), false, false, 0);
    vb.pack_start (vbox, true, true, 0);
    vb.pack_start (status_box, false, false, 0);
    status_box.pack_start (eb, false, false, 0);
    eb.add (score);
    eb.set_border_width (5);
    add (vb);
    
    set_title (_("Multiplication Puzzle"));
    
    new_puzzle ();
    
    destroy += Gtk.main_quit;
  }
  
  Gtk.Action solve_action;
  Gtk.Widget setup_menu ()
  {
    var action_group = new Gtk.ActionGroup ("gmult-actions");
    
    var action = new Gtk.Action ("GameMenuAction", _("_Game"), null, null);
    action_group.add_action (action);
    
    action = new Gtk.Action ("NewAction", null, null, Gtk.STOCK_NEW);
    action.activate += on_menu_game_new;
    action_group.add_action_with_accel (action, "<control>N");
    
    action = new Gtk.Action ("SolveAction", _("_Solve"), null, null);
    action.activate += on_menu_game_solve;
    action_group.add_action (action);
    this.solve_action = action;
    
    action = new Gtk.Action ("CloseAction", null, null, Gtk.STOCK_CLOSE);
    action.activate += on_menu_game_close;
    action_group.add_action_with_accel (action, "<control>W");
    
    action = new Gtk.Action ("HelpMenuAction", _("_Help"), null, null);
    action_group.add_action (action);
    
    action = new Gtk.Action ("AboutAction", null, null, Gtk.STOCK_ABOUT);
    action.activate += on_menu_help_about;
    action_group.add_action (action);
    
    var ui = "\
<ui>\
  <menubar>\
    <menu name=\"GameMenu\" action=\"GameMenuAction\">\
      <menuitem name=\"New\" action=\"NewAction\" />\
      <separator />\
      <menuitem name=\"Solve\" action=\"SolveAction\" />\
      <separator />\
      <menuitem name=\"Close\" action=\"CloseAction\" />\
    </menu>\
    <menu name=\"HelpMenu\" action=\"HelpMenuAction\">\
      <menuitem name=\"About\" action=\"AboutAction\"/>\
    </menu>\
  </menubar>\
</ui>";

    var manager = new Gtk.UIManager ();
    try {
    manager.add_ui_from_string (ui, ui.size ());
    } catch (Error e)  {
      error ("Internal error: bad ui string.\n");
    }
    manager.insert_action_group (action_group, 0);
    add_accel_group (manager.get_accel_group ());
    
    var menu_item = (Gtk.Item)manager.get_widget ("/ui/menubar/GameMenu/New");
    menu_item.select += on_menu_select_new;
    menu_item.deselect += on_menu_deselect;
    
    menu_item = (Gtk.Item)manager.get_widget ("/ui/menubar/GameMenu/Solve");
    menu_item.select += on_menu_select_solve;
    menu_item.deselect += on_menu_deselect;
    
    menu_item = (Gtk.Item)manager.get_widget ("/ui/menubar/GameMenu/Close");
    menu_item.select += on_menu_select_close;
    menu_item.deselect += on_menu_deselect;
    
    menu_item = (Gtk.Item)manager.get_widget ("/ui/menubar/HelpMenu/About");
    menu_item.select += on_menu_select_about;
    menu_item.deselect += on_menu_deselect;
    
    return manager.get_widget ("/ui/menubar");
  }

  private Gtk.Widget canvas;
  private Gtk.VBox vbox;
  private Gtk.Label score;
  private Gtk.HBox status_box;
  
  private void new_puzzle ()
  {
    if (canvas != null)
      vbox.remove (canvas);
    
    if (status != null)
      status_box.remove (status);
    
    // easiest way to reset the statusbar
    status = new Gtk.Statusbar ();
    status.set_has_resize_grip (false);
    status_box.pack_start (status, true, true, 0);
    status_box.reorder_child (status, 0);
    
    puzzle = new MultPuzzle (3, 2);
    
    canvas = new Canvas (this);
    vbox.pack_start (canvas, true, true, 0);
    
    Gdk.Geometry geom = Gdk.Geometry();
    geom.min_aspect = ASPECT;
    geom.max_aspect = ASPECT;
    set_geometry_hints(canvas, geom, Gdk.WindowHints.ASPECT);
    
    cheated = false;
    
    show_all ();
    update_score ();
    on_puzzle_change(puzzle);
    
    if (status_timeout_id != 0)
      Source.remove (status_timeout_id);
    status_timeout_id = 0;
    
    puzzle.guessed += on_puzzle_guess;
    puzzle.changed += on_puzzle_change;
  }
  
  private uint status_timeout_id;
  private void on_puzzle_guess (MultPuzzle p, int digit, MultPuzzleChar letter, MultPuzzleGuessStatus response)
  {
    update_score ();
    
    var context_id = status.get_context_id ("guess-feedback");
    
    // clear any previous message
    status.pop (context_id);
    if (status_timeout_id != 0)
      Source.remove (status_timeout_id);
    status_timeout_id = 0;
    
    if (puzzle.is_done)
    {
      status.push (context_id, _("Congratulations!"));
    }
    else
    {
      string message_pattern = null;
      
      switch (response)
      {
      case MultPuzzleGuessStatus.WRONG:
        // Translators: %1$s is the letter, %2$s is the digit
        message_pattern = _("Incorrect — %1$s is not %2$s");
        break;
      case MultPuzzleGuessStatus.CORRECT:
        // Translators: %1$s is the letter, %2$s is the digit
        message_pattern = _("Correct — %1$s is %2$s");
        break;
      default:
      case MultPuzzleGuessStatus.KNOWN:
      case MultPuzzleGuessStatus.INVALID:
        // shouldn't happen
        break;
      }
      
      if (message_pattern != null) {
        var letter_str = "%c".printf ((char)letter);
        var digit_str = "%i".printf (digit);
        status.push (context_id, message_pattern.printf (_(letter_str), _(digit_str)));
        status_timeout_id = Timeout.add_seconds (5, pop_guess_feedback);
      }
    }
  }
  
  private void on_puzzle_change (MultPuzzle p)
  {
    solve_action.set_sensitive (!p.is_done);
  }

  private bool pop_guess_feedback ()
  {
    var context_id = status.get_context_id ("guess-feedback");
    status.pop (context_id);
    status_timeout_id = 0;
    return false;
  }

  private void update_score ()
  {
    var wrong = puzzle.wrong_guesses;
    
    // Translators: The '%s' is an optional asterisk that appears when the
    // user has 'cheated' by asking for the solution to the puzzle.
    var text = _("%i Wrong%s").printf (wrong, cheated ? "*" : "");
    
    score.set_text (text);
  }
  
  private void on_menu_deselect ()
  {
    var context_id = status.get_context_id ("menu-hover");
     status.pop (context_id);
  }
  
  private void on_menu_game_close ()
  {
    destroy ();
  }
  
  private void on_menu_select_close ()
  {
    var context_id = status.get_context_id ("menu-hover");
     status.push (context_id, _("Close Multiplication Puzzle"));
  }
  
  private void on_menu_game_new ()
  {
    new_puzzle ();
  }
  
  private void on_menu_select_new ()
  {
    var context_id = status.get_context_id ("menu-hover");
     status.push (context_id, _("Start a new game"));
  }
  
  private void on_menu_game_solve ()
  {
    cheated = true;
    
    puzzle.solve ();
    
    var context_id = status.get_context_id ("guess-feedback");
    
    // Clear any previous message
    status.pop (context_id);
    if (status_timeout_id != 0)
      Source.remove (status_timeout_id);
    status_timeout_id = 0;
    
    status.push (context_id, _("Puzzle solved"));
    
    update_score ();
  }
  
  private void on_menu_select_solve ()
  {
    var context_id = status.get_context_id ("menu-hover");
     status.push (context_id, _("Solve this game"));
  }
  
  private void handle_about_uri (Gtk.AboutDialog about, string link)
  {
    AppLaunchContext context = new AppLaunchContext ();
    
    try {
      AppInfo.launch_default_for_uri (link, context);
    } catch (Error er) {
      Gtk.MessageDialog dlg = new Gtk.MessageDialog.with_markup (this, Gtk.DialogFlags.DESTROY_WITH_PARENT | Gtk.DialogFlags.MODAL, Gtk.MessageType.ERROR, Gtk.ButtonsType.OK, "<b>%s</b>", _("Could not display %s").printf (link));
      dlg.run ();
      dlg.destroy ();
    }
  }
  
  private void handle_about_mail (Gtk.AboutDialog about, string link)
  {
    handle_about_uri (about, "mailto:%s".printf (link));
  }
  
  // These need to be class-wide to prevent an odd compiler syntax error.
  private const string[] authors = {"Michael Terry <mike@mterry.name>",
                                    null};
  private const string[] artists = {"Eugenia Loli-Queru <eloli@hotmail.com>",
                                    "Michael Terry <mike@mterry.name>",
                                    null};
  
  private Gtk.AboutDialog about { get; set; default = null; }
  private void on_menu_help_about ()
  {
    if (about != null)
    {
      about.present ();
      return;
    }
    
    about = new Gtk.AboutDialog ();
    about.set_email_hook (handle_about_mail, null);
    about.set_url_hook (handle_about_uri, null);
    about.title = _("About Multiplication Puzzle");
    about.authors = authors;
    about.artists = artists;
    about.translator_credits = _("translator-credits");
    about.logo_icon_name = Config.PACKAGE;
    about.version = Config.VERSION;
    about.copyright = "© 2003—2008 Michael Terry";
    about.website = "http://mterry.name/gmult/";
    about.license = "%s\n\n%s\n\n%s".printf (
      _("This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 3 of the License, or (at your option) any later version."),
      _("This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details."),
      _("You should have received a copy of the GNU General Public License along with this program; if not, write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA."));
    about.wrap_license = true;
    
    about.set_transient_for (this);
    about.response += (dlg, resp) => {dlg.destroy (); about = null;};
    
    about.show ();
  }
  
  private void on_menu_select_about ()
  {
    var context_id = status.get_context_id ("menu-hover");
     status.push (context_id, _("About Multiplication Puzzle"));
  }
  
  public static int main(string[] args)
  {
    GLib.Intl.textdomain (Config.GETTEXT_PACKAGE);
    GLib.Intl.bindtextdomain (Config.GETTEXT_PACKAGE, Config.LOCALE_DIR);
    GLib.Intl.bind_textdomain_codeset (Config.GETTEXT_PACKAGE, "UTF-8");
    
    OptionContext context = new OptionContext ("");
    context.add_group (Gtk.get_option_group (false)); // allow console use
    try {
      context.parse (ref args);
    } catch (Error e) {
      printerr ("%s\n\n%s", e.message, context.get_help (true, null));
      return 1;
    }
    
    Gtk.init (ref args); // to open display ('cause we passed false above)
    
    GLib.Environment.set_application_name (_("Multiplication Puzzle"));
    Gtk.IconTheme.get_default ().append_search_path (Config.THEME_DIR);
    Gtk.Window.set_default_icon_name (Config.PACKAGE);
    
    var window = new GtkMult ();
    window.show_all();
    Gtk.main ();
    
    return 0;
  }
}

